// $Id: Dump.java 25247 2017-04-10 18:03:59Z mike $

import org.faceless.pdf2.*;
import java.io.*;
import java.awt.color.ICC_Profile;
import java.util.*;
import java.net.URL;
import java.net.MalformedURLException;
import java.awt.geom.*;
import java.awt.Shape;
import org.w3c.dom.*;
import javax.xml.transform.*;
import javax.xml.transform.stream.*;
import javax.xml.transform.dom.*;
import java.text.*;
import java.security.*;
import java.security.cert.X509Certificate;
import java.security.cert.X509CRL;

/**
 * Usage: java Dump [--out <filename>] [--profile full | basic] [--structure] [--keystore <keystore>]
 *                  [--password <password>] [--provider <providername>] [--laxssl] <input>...
 *
 * This example takes a PDF file and dumps out as much information as possible
 * about it. It's very complete so may display a lot of information!
* <input> may be a File, a URL, or "-" to read the PDF from stdin.
 *
 * The command line options are:
 * --out <filename>             Redirects output to a file. The file will be in UTF-8
 *                              format, unlike output to System.out which is escaped.
 * --profile full               Extract the Full OutputProfile rather than the basic one.
 * --structure                  Dump the PDF structure, if available. Implies "--profile full"
 * --keystore <keystore>        Specify the keystore (for use with public key encrypted docs,
 *                              or for verifying signatures against a list of trusted roots.
 * --password <password>        Specify the open password for the document.
 * --provider <provider>        Use the specified Security Provider. Typical values for
 *                              this would be "BC" to use the BouncyCastle provider, "IAIK"
 *                              to use the IAIK provider, "SUN" for the default provider
 *                              supplied with Java, or the class name of a Provider.
 * --laxssl                     Don't check certificates when loading a PDF from an HTTPS URL
 */
public class Dump
{
    public static void main(String[] args) throws Exception {
        if (args.length == 0) {
            System.err.println("USAGE: java Dump [--out <filename>] [--keystore keystore] [--profile full | basic] [--password <password>] [--provider <providername>] [--laxssl] <file|url|-> ...\n");
            System.exit(1);
        }

        PrintWriter SYSTEMOUT = new PrintWriter(new OutputStreamWriter(System.out, "UTF-8"));
        PrintWriter out = SYSTEMOUT;
        String filename = null, password = null, keystore = null;
        boolean fullprofile = false, structure = false;
        for (int i=0;i<args.length;i++) {
            if (args[i].equals("--password")) {
                password = args[++i];
            } else if (args[i].equals("--structure")) {
                structure = true;
            } else if (args[i].equals("--keystore")) {
                keystore = args[++i];
            } else if (args[i].equals("--laxssl")) {
                laxssl();
            } else if (args[i].equals("--profile")) {
                fullprofile = args[++i].equals("full");
            } else if (args[i].equals("--provider")) {
                String provname = args[++i];
                if (provname.equalsIgnoreCase("BC")) {
                    provname = "org.bouncycastle.jce.provider.BouncyCastleProvider";
                } else if (provname.equalsIgnoreCase("IAIK")) {
                    provname = "iaik.security.provider.IAIK";
                }
                Provider provider = Security.getProvider(provname);
                if (provider == null) {
                    Class cl = null;
                    try {
                        cl = Class.forName(provname);
                    } catch (ClassNotFoundException e) {}
                    if (cl != null && Provider.class.isAssignableFrom(cl)) {
                        provider = (Provider)cl.newInstance();
                    } else {
                        throw new IllegalArgumentException("No Security Provider \"" + provname + "\"");
                    }
                }
                Security.insertProviderAt(provider, 1);
            } else if (args[i].equals("--out")) {
                if (out != SYSTEMOUT) {
                    out.close();
                    System.out.println("Created \"" + filename + "\"");
                }
                filename = args[++i];
                out = new PrintWriter(new OutputStreamWriter(new FileOutputStream(filename), "UTF-8"), true);
            } else {
                EncryptionHandler handler = null;
                KeyStore store = null;
                if (keystore != null) {
                    store = KeyStore.getInstance("JKS");
                    FileInputStream kin = new FileInputStream(new File(keystore));
                    store.load(kin, null);
                    kin.close();
                }
                if (store != null && password != null) {
                    handler = new PublicKeyEncryptionHandler(store, null, password.toCharArray());
                } else if (password != null) {
                    handler = new StandardEncryptionHandler();
                    ((StandardEncryptionHandler)handler).setUserPassword(password);
                }
                Dump dump = new Dump(args[i], handler, out, fullprofile || structure);
                dump.keystore = store;
                dump.unicode = filename != null;
                dump.structure = structure;
                dump.go();
                out.flush();
            }
        }

        if (out != SYSTEMOUT) {
            out.close();
            System.out.println("Created \"" + filename + "\"");
        }
    }

    private static void laxssl() throws GeneralSecurityException {
        javax.net.ssl.SSLContext sc = javax.net.ssl.SSLContext.getInstance("SSL");
        sc.init(null, new javax.net.ssl.TrustManager[] { new javax.net.ssl.X509TrustManager() {
            public void checkClientTrusted(java.security.cert.X509Certificate[] chain, String auth) { }
            public void checkServerTrusted(java.security.cert.X509Certificate[] chain, String auth) { }
            public java.security.cert.X509Certificate[] getAcceptedIssuers() { return null; }
        } }, null);
        javax.net.ssl.HttpsURLConnection.setDefaultSSLSocketFactory(sc.getSocketFactory());
        javax.net.ssl.HttpsURLConnection.setDefaultHostnameVerifier(new javax.net.ssl.HostnameVerifier() {
            public boolean verify(String urlHostname, javax.net.ssl.SSLSession session) { return true; }
        });
    }


    //--------------------------------------------------------------

    private final PDF pdf;
    private final PDFParser pdfparser;
    private final PrintWriter out;
    private boolean unicode, structure;
    private KeyStore keystore;
    private final DateFormat dateformat = new SimpleDateFormat("dd MMM yyyy HH:mm:ss ZZZZ");

    Dump(String file, EncryptionHandler handler, PrintWriter out, boolean fullprofile) throws IOException, PasswordException {
        out.println("DUMPING \"" + file + "\" with PDF Library " + PDF.VERSION + "/" + System.getProperty("java.vendor") + " Java " + System.getProperty("java.version") + "/" + System.getProperty("os.name"));
        PDFReader reader = new PDFReader();
        reader.addEncryptionHandler(handler);
        try {
            if (file.equals("-")) {
                reader.setSource(System.in);
            } else {
                try {
                    URL url = new URL(file);
                    reader.setSource(url);
                } catch (MalformedURLException e) {
                    reader.setSource(new File(file));
                }
            }
        } catch (IOException e) {
            if (e.getCause() instanceof PasswordException) {
                throw (PasswordException)e.getCause();
            } else {
                throw e;
            }
        }
        reader.load();
        this.pdf = new PDF(reader);
        this.pdfparser = fullprofile ? new PDFParser(pdf) : null;
        this.out = out;
        out.println("---------------------------------------------------------------");
    }

    void go() throws Exception {
        Printer printer = new Printer(out);
        getInfoDictionary(printer);
        getOptions(printer);
        OutputProfile profile = getOutputProfile(printer);
        getOutputIntent(printer, profile, "GTS_PDFX", "PDF/X");
        getOutputIntent(printer, profile, "GTS_PDFA1", "PDF/A");
        if (pdfparser != null) {
            if (profile.isSet(OutputProfile.Feature.InfoMeetsPDFA1a_2005)) {
                getProfileValidation(printer, OutputProfile.PDFA1a_2005, "PDF/A-1a:2005", profile);
            } else if (profile.isSet(OutputProfile.Feature.InfoMeetsPDFA1b_2005)) {
                getProfileValidation(printer, OutputProfile.PDFA1b_2005, "PDF/A-1b:2005", profile);
            }
            if (profile.isSet(OutputProfile.Feature.InfoMeetsPDFA2a)) {
                getProfileValidation(printer, OutputProfile.PDFA2a, "PDF/A-2a", profile);
            } else if (profile.isSet(OutputProfile.Feature.InfoMeetsPDFA2b)) {
                getProfileValidation(printer, OutputProfile.PDFA2b, "PDF/A-2b", profile);
            }
            if (profile.isSet(OutputProfile.Feature.InfoMeetsPDFA3a)) {
                getProfileValidation(printer, OutputProfile.PDFA3a, "PDF/A-3a", profile);
            } else if (profile.isSet(OutputProfile.Feature.InfoMeetsPDFA3a)) {
                getProfileValidation(printer, OutputProfile.PDFA3b, "PDF/A-3b", profile);
            }
            if (profile.isSet(OutputProfile.Feature.InfoMeetsPDFX1a_2001)) {
                getProfileValidation(printer, OutputProfile.PDFX1a_2001, "PDF/X-1a:2001", profile);
            } else if (profile.isSet(OutputProfile.Feature.InfoMeetsPDFX1a_2003)) {
                getProfileValidation(printer, OutputProfile.PDFX1a_2003, "PDF/X-1a:2003", profile);
            }
            if (profile.isSet(OutputProfile.Feature.InfoMeetsPDFX3_2002)) {
                getProfileValidation(printer, OutputProfile.PDFX3_2002, "PDF/X-3:2002", profile);
            }
            if (profile.isSet(OutputProfile.Feature.InfoMeetsPDFX3_2003)) {
                getProfileValidation(printer, OutputProfile.PDFX3_2003, "PDF/X-3:2003", profile);
            }
            if (profile.isSet(OutputProfile.Feature.InfoMeetsPDFUA1)) {
                getProfileValidation(printer, OutputProfile.PDFUA1, "PDF/UA-1", profile);
            }
        }
        getEncryption(printer);
        getJavaScript(printer);
        getDocumentActions(printer);
        getMetaData(pdf.getMetaData(), printer);
        getBookmarks(printer);
        getNamedActions(printer);
        getEmbeddedFiles(printer);
        if (structure) {
            getStructure(printer);
        }
        getPages(printer);
        getXFA(printer);
        getForm(printer);
    }

    void getProfileValidation(Printer out, OutputProfile target, String name, OutputProfile profile) {
        out.startSection(name + " Compliance");
        OutputProfile.Feature[] mismatch = profile.isCompatibleWith(target);
        if (mismatch == null) {
            out.println("Passed");
        } else {
            out.println("Failed because");
            for (int i=0;i<mismatch.length;i++) {
                if (profile.isSet(mismatch[i])) {
                    out.println("* \"" + mismatch[i] + "\" is set but disallowed");
                } else {
                    out.println("* \"" + mismatch[i] + "\" is required but missing");
                }
            }
        }
        out.endSection();
    }

    @SuppressWarnings("unchecked")
    void getStructure(Printer out) {
        Document doc = pdfparser.getStructureTree();
        if (doc != null) {
            out.startSection("Structure");
            out.print("");
            try {
                Transformer transformer = TransformerFactory.newInstance().newTransformer();
                transformer.setOutputProperty(OutputKeys.INDENT, "yes");
                transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
                transformer.transform(new DOMSource(doc), new StreamResult(out.writer));
            } catch (Exception e) {
                e.printStackTrace();
            }
            out.endSection();
        }
    }

    @SuppressWarnings("unchecked")
    void getOptions(Printer out) {
        final String[] keys = { "view.fullscreen", "view.displaydoctitle", "view.hidetoolbar", "view.hidemenubar", "view.hidewindowui", "view.fitwindow", "view.centerwindow", "pagelayout", "pagemode", "view.area", "view.clip", "print.area", "print.clip", "print.scaling", "print.duplex", "print.matchtraysize", "print.numcopies", "print.pagerange" };

        out.startSection("Options");
        for (int i=0;i<keys.length;i++) {
            String key = keys[i];
            Object value = pdf.getOption(key);
            if (value != null) {
                if (key == "print.pagerange") {
                    List pagelist = (List)value;                // a List of PDFPages
                    for (int j=0;j<pagelist.size();j++) {
                        PDFPage page = (PDFPage)pagelist.get(j);
                        pagelist.set(j, Integer.valueOf(page.getPageNumber()));
                    }
                }
                out.println(keys[i] + ":\t" + value);
            }
        }
        out.endSection();
    }

    void getEncryption(Printer out) {
        out.startSection("Encryption");
        EncryptionHandler handler = pdf.getEncryptionHandler();
        if (handler != null) {
            if (handler instanceof StandardEncryptionHandler || handler instanceof PublicKeyEncryptionHandler) {
                int change, print, extract;
                if (handler instanceof StandardEncryptionHandler) {
                    int version = ((StandardEncryptionHandler)handler).getVersion();
                    out.println("Standard PDF Encryption version " + version + ": " + ((StandardEncryptionHandler)handler).getDescription());
                    change = ((StandardEncryptionHandler)handler).getChange();
                    print = ((StandardEncryptionHandler)handler).getPrint();
                    extract = ((StandardEncryptionHandler)handler).getExtract();
                } else {
                    out.println("Public Key Encryption");
                    change = ((PublicKeyEncryptionHandler)handler).getChange();
                    print = ((PublicKeyEncryptionHandler)handler).getPrint();
                    extract = ((PublicKeyEncryptionHandler)handler).getExtract();
                }
                boolean change_forms = (change & StandardEncryptionHandler.CHANGE_FORMS) == StandardEncryptionHandler.CHANGE_FORMS;
                boolean change_layout = (change & StandardEncryptionHandler.CHANGE_LAYOUT) == StandardEncryptionHandler.CHANGE_LAYOUT;
                boolean change_annots = (change & StandardEncryptionHandler.CHANGE_ANNOTATIONS) == StandardEncryptionHandler.CHANGE_ANNOTATIONS;
                boolean change_none = !(change_forms | change_layout | change_annots);
                out.println("CanChange:\t" + (change_none?"None ":"")+(change_layout?"Assembly ":"")+(change_forms?"Forms ":"")+(change_annots?"Comments ":"") + " " + change);
                out.println("CanPrint:\t" + (print == StandardEncryptionHandler.PRINT_NONE ? "None" : print == StandardEncryptionHandler.PRINT_LOWRES ? "Low-res" : "High-res") + " " + print);
                out.println("CanExtract:\t" + (extract == StandardEncryptionHandler.EXTRACT_NONE ? "None" : extract == StandardEncryptionHandler.EXTRACT_ACCESSIBILITY ? "Accessibility" : "All") + " " + extract);
            } else {
                out.println("Non-standard Encryption Handler: " + handler.getFilterName());
            }
        }
        out.endSection();
    }

    OutputProfile getOutputProfile(Printer out) {
        out.startSection("Output Profile");
        long time1 = System.currentTimeMillis();
        OutputProfiler profiler;
        if (pdfparser != null) {
            profiler = new OutputProfiler(pdfparser);
        } else {
            profiler = new OutputProfiler(pdf);
        }
        OutputProfile profile = profiler.getProfile();
        long time2 = System.currentTimeMillis();
        for (int i=0;i<OutputProfile.Feature.ALL.length;i++) {
            OutputProfile.Feature feature = OutputProfile.Feature.ALL[i];
            if (profile.isSet(feature)) {
                out.println("\"" + feature + "\" is set");
            }
        }
        Set<OutputProfile.FontInfo> fonts = profile.getFontInfo();
        if (fonts != null) {
            out.startSection("Fonts");
            for (Iterator<OutputProfile.FontInfo> i = fonts.iterator();i.hasNext();) {
                OutputProfile.FontInfo fontinfo = i.next();
                out.println(fontinfo.getBaseName() + ": " + (fontinfo.isCIDFont() ? "CID ":" Simple ")+fontinfo.getType()+(fontinfo.isEmbedded() ? " (Embedded)" : ""));
            }
            out.endSection();
        }
        if (pdfparser != null) {
            out.println("Output profiling took " + (time2-time1) + "ms");
        }
        out.endSection();
        return profile;
    }

    void getOutputIntent(Printer out, OutputProfile profile, String type, String name) {
        if (profile.getOutputConditionIdentifier(type) != null) {
            out.startSection(name + " Output Intent");
            out.println("Identifier:\t" + profile.getOutputConditionIdentifier(type));
            out.println("Condition:\t" + profile.getOutputCondition(type));
            out.println("Registry:\t" + profile.getOutputConditionRegistry(type));
            out.println("Info:    \t" + profile.getOutputIntentInfo(type));
            ICC_Profile icc = profile.getOutputIntentDestinationProfile(type);
            if (icc == null) {
                out.println("Profile: \tnot embedded");
            } else {
                out.println("Profile: \tembedded (" + icc.getNumComponents() + " components)");
            }
            out.endSection();
        }
    }

    void getJavaScript(Printer out) {
        out.startSection("JavaScript");
        String js = pdf.getJavaScript();
        if (js != null) {
            out.println(js);
        }
        out.endSection();
    }

    void getBookmarks(Printer out) {
        out.startSection("Bookmarks");
        getBookmarks(out, pdf.getBookmarks(), "... ");
        out.endSection();
    }

    private void getBookmarks(Printer out, List<PDFBookmark> l, String indent) {
        for (int i=0;l != null && i<l.size();i++) {
            PDFBookmark b = l.get(i);
            if (b != null) {
                out.println(indent+b.getName() + ":\t" + getAction(b.getAction(), null, null));
                getBookmarks(out, b.getBookmarks(), "..." + indent);
            }
        }
    }

    private String getAction(PDFAction action, String name, Printer out) {
        boolean multiline = out != null;
        StringWriter sw = null;
        if (out == null) {
            out = new Printer(new PrintWriter(sw = new StringWriter()));
        }

        if (action != null) {
            if (multiline) {
                out.startSection(name);
            }
            while (action != null) {
                String type = action.getType();
                if (type.equals("FormJavaScript")) {
                    if (multiline) {
                        out.println(action.getJavaScript());
                    } else {
                        out.println("JavaScript: " + action.getJavaScript().length() + " chars");
                    }
                } else if (type.equals("GoToR")) {
                    out.println("Remote: " + action.getRemoteFilename() + " page " + action.getRemotePageNumber());
                } else if (type.equals("FormReset") || type.equals("FormSubmit")) {
                    String desc = type + ":";
                    if (type.equals("FormSubmit")) {
                        desc += " URL=" + action.getURL();
                        int method = action.getFormSubmitMethod();
                        if (method == PDFAction.METHOD_HTTP_POST) {
                            desc += " method=POST";
                        } else if (method == PDFAction.METHOD_HTTP_IMAGEMAP_POST) {
                            desc += " method=POST(Imagemap)";
                        } else if (method == PDFAction.METHOD_FDF) {
                            desc += " method=FDF";
                        } else if (method == PDFAction.METHOD_XML) {
                            desc += " method=XML";
                        } else if (method == PDFAction.METHOD_FDF_WITH_ALL_ANNOTATIONS) {
                            desc += " method=FDF(include all annotations)";
                        } else if (method == PDFAction.METHOD_XML_WITH_ALL_ANNOTATIONS) {
                            desc += " method=XML(include all annotations)";
                        } else if (method == PDFAction.METHOD_FDF_WITH_MY_ANNOTATIONS) {
                            desc += " method=FDF(include my annotations)";
                        } else if (method == PDFAction.METHOD_PDF) {
                            desc += " method=PDF";
                        } else {
                            desc += " method=Unknown!";
                        }
                        int flags = action.getFormSubmitFlags();
                        if ((flags & PDFAction.METHOD_FLAG_EMPTYFIELDS) != 0) {
                            desc +=" emptyfields=true";
                        }
                        if ((flags & PDFAction.METHOD_FLAG_CANONICALDATES) != 0) {
                            desc +=" canonicaldates=true";
                        }
                        desc += " flags=0x" + Integer.toHexString(flags);
                    }
                    Collection<FormElement> c = action.getFormSubmitFields();
                    String fieldnames;
                    if (c.size() == pdf.getForm().getElements().size()) {
                        desc += " fields=<all>";
                    } else {
                        List<String> l = new ArrayList<String>();
                        for (Iterator<FormElement> i = c.iterator();i.hasNext();) {
                            FormElement e = i.next();
                            l.add(e.getForm().getName(e));
                        }
                        desc += " fields=" + l;
                    }
                    out.println(desc);
                } else if (type.startsWith("Named")) {
                    out.println(type);
                } else if (type.equals("URL")) {
                    out.println("URL: " + action.getURL());
                } else if (type.equals("Launch")) {
                    out.println("Launch: " + action.getRemoteFilename());
                } else if (type.equals("ShowWidget") || type.equals("HideWidget")) {
                    WidgetAnnotation annot = action.getAnnotation();
                    FormElement field = annot == null ? null : annot.getField();
                    String fieldname = field == null ? null : field.getForm().getName(field);
                    out.println(type + " " + fieldname + ":" + (field == null?0:field.getAnnotations().indexOf(annot)));
                } else if (type.startsWith("GoTo")) {
                    int pagenumber = action.getPageNumber();
                    // To better test linearization, optionally don't get the coordinates of
                    // the action on the page unless the page is already loaded.
                    String stringcoords = null;
                    if (pagenumber >= 0 /*&& pdf.getLoadState(pagenumber) == null*/) {
                        float[] coords = action.getGoToCoordinates();
                        stringcoords = null;
                        if (coords == null) {
                            stringcoords = null;
                        } else if (coords.length == 1) {
                            stringcoords = "[" + fmt(coords[0]) + "]";
                        } else if (coords.length == 2) {
                            stringcoords = "[" + fmt(coords[0]) + " " + fmt(coords[1]) + "]";
                        } else if (coords.length == 3) {
                            stringcoords = "[" + fmt(coords[0]) + " " + fmt(coords[1]) + " " + fmt(coords[2]) + "]";
                        } else if (coords.length == 4) {
                            stringcoords = "[" + fmt(coords[0]) + " " + fmt(coords[1]) + " " + fmt(coords[2]) + " " + fmt(coords[3]) + "]";
                        }
                    }
                    out.println(type + ": page=" + (pagenumber < 0 ? "nopage" : Integer.toString(pagenumber+1)) + " coords=" + stringcoords);
                } else {
                    out.println(type);
                }
                action = action.getNext();
            }
            if (multiline) {
                out.endSection();
            }
        }
        if (sw != null) {
            String s = sw.toString();
            if (s.length() == 0) {
                s = null;
            } else {
                s = s.substring(0, s.length()-1);
                if (s.indexOf("\n") >= 0) {
                    s = "[" + s.replaceAll("\n", ", ") + "]";
                }
            }
            return s;
        } else {
            return null;
        }
    }

    void getAnnotation(PDFAnnotation annot, String title, Printer out) throws IOException {
        out.startSection(title);
        out.println("Type:\t" + annot.getType()+" ("+annot.getClass().getName()+")");
        float[] r = annot.getRectangle();
        if (r != null) {
            out.println("Rect:\t" + fmt(r[0]) + ", " + fmt(r[1]) + " - " + fmt(r[2]) + ", " + fmt(r[3]));
        }
        if (annot.getColor() != null) {
            out.println("Color :\t" + annot.getColor());
        }
        if (annot.getUniqueID() != null) {
            out.println("Unique ID:\t" + annot.getUniqueID());
        }
        if (annot.getModifyDate() != null) {
            out.println("Modified:\t" + dateformat.format(annot.getModifyDate().getTime()));
        }
        if (annot.getCreationDate() != null) {
            out.println("Created:\t" + dateformat.format(annot.getCreationDate().getTime()));
        }
        if (annot.isReadOnly()) {
            out.println("Read-Only:\ttrue");
        }
        if (annot.isPositionLocked() || annot.isContentLocked()) {
            out.println("Locked:\t" + (annot.isPositionLocked() ? "position ":"")+(annot.isContentLocked()?"content":""));
        }
        if (!annot.isVisible()) {
            out.println("Invisible:\ttrue");
        }
        if (!annot.isPrintable()) {
            out.println("Printable:\tfalse");
        }
        if (annot instanceof AnnotationFile) {
            getEmbeddedFile("File", ((AnnotationFile)annot).getFile(), out);
        }
        if (annot instanceof WidgetAnnotation) {
            WidgetAnnotation widget = (WidgetAnnotation)annot;
            out.println("Field:\t" + pdf.getForm().getName(widget.getField()));
            out.println("Styles:\t" + widget.getTextStyle()+", "+widget.getBackgroundStyle());
            for (int i=0;i<Event.ALL.length;i++) {
                PDFAction act = widget.getAction(Event.ALL[i]);
                getAction(act, Event.ALL[i].toString(), out);
            }
        } else {
            if (annot instanceof AnnotationLink) {
                getAction(((AnnotationLink)annot).getAction(), "Action", out);
                float[] corners = ((AnnotationLink)annot).getCorners();
                if (corners != null) {
                    StringBuffer sb = new StringBuffer();
                    for (int i=0;i<corners.length;i+=2) {
                        sb.append(fmt.format(corners[i]) + "," + fmt.format(corners[i+1]) + " ");
                        if (i%8 == 6) {
                            sb.append("  ");
                        }
                    }
                    out.println("Corners:\t" + sb);
                }
            } else if (annot instanceof AnnotationMarkup) {
                float[] corners = ((AnnotationMarkup)annot).getCorners();
                if (corners != null) {
                    StringBuffer sb = new StringBuffer();
                    for (int i=0;i<corners.length;i+=2) {
                        sb.append(fmt.format(corners[i]) + "," + fmt.format(corners[i+1]) + " ");
                        if (i%8 == 6) {
                            sb.append("  ");
                        }
                    }
                    out.println("Corners:\t" + sb);
                }
            } else if (annot instanceof AnnotationShape) {
                Shape shape = ((AnnotationShape)annot).getShape();
                out.println("Shape:\t" + fmt(shape));
                out.println("Style:\t" + ((AnnotationShape)annot).getStyle());
            }
            if (annot.getSubject() != null) {
                out.println("Subject:\t" + annot.getSubject());
            }
            if (annot.getAuthor() != null) {
                out.println("Author:\t" + annot.getAuthor());
            }
        }
        getMetaData(annot.getMetaData(), out);
        if (annot.getContents() != null) {
            out.println("");
            out.println(fmt(annot.getContents().trim()));
        }
        List<PDFAnnotation> l = annot.getReviews();
        for (int i=0;i<l.size();i++) {
            PDFAnnotation subannot = l.get(i);
            if (subannot instanceof AnnotationNote) {
                AnnotationNote note = (AnnotationNote)subannot;
                out.println(" * " + note.getStatus() + " set by " + note.getAuthor() + " at " + dateformat.format(note.getModifyDate().getTime()));
            } else {
                getAnnotation(subannot, "Review " + (i+1) + "/" + l.size(), out);
            }
        }
        l = annot.getReplies();
        for (int i=0;i<l.size();i++) {
            getAnnotation((PDFAnnotation)l.get(i), "Reply " + (i+1) + "/" + l.size(), out);
        }
        out.endSection();
    }

    void getInfoDictionary(Printer out) {
        out.startSection("Info Dictionary");
        out.println("DocumentID:\t" + pdf.getDocumentID(true) + " " + pdf.getDocumentID(false));
        out.println("Revisions:\t" + pdf.getNumberOfRevisions());
        Map<String,Object> info = pdf.getInfo();
        for (Iterator<Map.Entry<String,Object>> i = info.entrySet().iterator();i.hasNext();) {
            Map.Entry<String,Object> e = i.next();
            String key = e.getKey();
            Object val = e.getValue();
            String sval;
            if (val instanceof Calendar && key.charAt(0)=='_' && info.containsKey(key.substring(1))) {
                key = key.substring(1);
                dateformat.setTimeZone(((Calendar)val).getTimeZone());
                sval = dateformat.format(((Calendar)val).getTime());
                dateformat.setTimeZone(TimeZone.getDefault());
            } else if (val instanceof Date && info.containsKey("_"+key)) {
                sval = null;
            } else {
                sval = val.toString();
            }
            if (sval != null) {
                out.println(key + ":\t" + fmt(sval));
            }
        }
        if (pdf.getLocale() != null) {
            out.println("Locale:\t" + pdf.getLocale());
        }
        out.endSection();
    }

    void getDocumentActions(Printer out) {
        out.startSection("Document Actions");
        for (int i=0;i<Event.ALL.length;i++) {
            PDFAction act = pdf.getAction(Event.ALL[i]);
            getAction(act, Event.ALL[i].toString(), out);
        }
        out.endSection();
    }

    void getMetaData(Reader metadata, Printer out) {
        out.startSection("MetaData");
        if (metadata != null) {
            StringWriter sb = new StringWriter();
            try {
                BufferedReader br = new BufferedReader(metadata);
                if (true) {     // Always pretty-print XMP metadata
                    Transformer transformer = TransformerFactory.newInstance().newTransformer();
                    transformer.setOutputProperty(OutputKeys.INDENT, "yes");
                    transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
                    transformer.transform(new StreamSource(br), new StreamResult(sb));
                } else {
                    String s;
                    while ((s=br.readLine()) != null) {
                        sb.append(s + "\n");
                    }
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
            out.println(sb.toString().trim());
        }
        out.endSection();
    }

    void getNamedActions(Printer out) {
        out.startSection("Named Actions");
        Map<String,PDFAction> actions = pdf.getNamedActions();
        for (Iterator<Map.Entry<String,PDFAction>> i = actions.entrySet().iterator();i.hasNext();) {
            Map.Entry<String,PDFAction> e = i.next();
            String name = e.getKey();
            PDFAction action = e.getValue();
            getAction(action, name, out);
        }
        out.endSection();
    }

    void getEmbeddedFiles(Printer out) throws IOException {
        out.startSection("Embedded Files");
        Map<String,EmbeddedFile> files = pdf.getEmbeddedFiles();
        for (Iterator<Map.Entry<String,EmbeddedFile>> i = files.entrySet().iterator();i.hasNext();) {
            Map.Entry<String,EmbeddedFile> e = i.next();
            String name = e.getKey();
            EmbeddedFile value = e.getValue();
            getEmbeddedFile(name, value, out);
        }
        out.endSection();
    }

    void getEmbeddedFile(String handle, EmbeddedFile file, Printer out) throws IOException {
        out.startSection(handle);
        if (file != null) {
            out.println("Name\t" + file.getName());
            if (file.getType() != null) {
                out.println("Type\t" + file.getType());
            }
            if (file.getModDate() != null) {
                out.println("Modified\t" + dateformat.format(file.getModDate()));
            }
            if (file.getCreationDate() != null) {
                out.println("Created\t" + dateformat.format(file.getCreationDate()));
            }
            if (file.getSize() != -1) {
                out.println("Size\t" + file.getSize());
            }
            if (file.getDescription() != null) {
                out.println("Description\t" + file.getDescription());
            }
            if (file.getPortfolioFolder() != null && !file.getPortfolioFolder().equals("/")) {
                out.println("Portfolio Folder\t" + file.getPortfolioFolder());
            }
            out.println("Checksum\t" + (file.isValid() ? "Probably OK" : "Definitely corrupt"));
        }
        out.endSection();
    }

    void getPages(Printer out) throws Exception {
        out.startSection("Pages");
        List<PDFPage> l = pdf.getPages();
        for (int i=0;i<l.size();i++) {
            PDFPage page = l.get(i);
            getPage(page, out);
        }
        out.endSection();
    }

    void getPage(PDFPage page, Printer out) throws Exception {
        String label = pdf.getPageLabel(page.getPageNumber() - 1);
        out.startSection("Page " + page.getPageNumber() + "/" + pdf.getNumberOfPages() + (label == null ? "" : "  (" + label + ")"));
        out.println("Page Size:\t" + page.getWidth() + " x " + page.getHeight());
        String[] boxes = { "CropBox", "ArtBox", "BleedBox", "TrimBox" };
        for (int i=0;i<boxes.length;i++) {
            float[] box = page.getBox(boxes[i]);
            if (box != null) {
                out.println(boxes[i] + ":\t" + fmt(box[0]) + ", " + fmt(box[1]) + " - " + fmt(box[2]) + ", " + fmt(box[3]));
            }
        }
        int rotate = page.getPageOrientation();
        if (rotate != 0) {
            out.println("Orientation:\t" + rotate);
        }
        for (int i=0;i<Event.ALL.length;i++) {
            PDFAction act = page.getAction(Event.ALL[i]);
            getAction(act, Event.ALL[i].toString(), out);
        }

        List<PDFAnnotation> annots = page.getAnnotations();
        for (int j=0;j<annots.size();j++) {
            PDFAnnotation annot = annots.get(j);
            if (annot.getInReplyTo() == null && !annot.getType().equals("Popup")) {
                getAnnotation(annot, "Annotation " + (j+1) + "/" + annots.size(), out);
            }
        }
        getMetaData(page.getMetaData(), out);
        out.endSection();
    }

    /**
     * Dump information on the XFA form. This method requires Java 1.4,
     * specifically the javax.xml.transform package. If you're still
     * running Java 1.3 you can comment it out and the call to it.
     */
    void getXFA(Printer out) throws Exception {
        out.startSection("XFA Form");
        InputStream in = pdf.getForm().getXFAElement(null);
        if (in != null) {
            StringWriter s = new StringWriter();
            Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
            transformer.transform(new StreamSource(in), new StreamResult(s));
            out.println(s.toString());
        }
        out.endSection();
    }

    void getForm(Printer out) throws Exception {
        out.startSection("Form");
        Map<String,FormElement> form = new TreeMap<String,FormElement>(pdf.getForm().getElements());

        for (Iterator<Map.Entry<String,FormElement>> i=form.entrySet().iterator(); i.hasNext();) {
            Map.Entry<String,FormElement> e = i.next();
            String name = e.getKey();
            FormElement el = e.getValue();
            if (el instanceof FormSignature) {
                getFormSignature((FormSignature)el, name, out);
            } else if (el instanceof FormText) {
                getFormText((FormText)el, name, out);
            } else if (el instanceof FormButton) {
                getFormButton((FormButton)el, name, out);
            } else if (el instanceof FormCheckbox || el instanceof FormRadioButton) {
                getFormRadioBox(el, name, out);
            } else if (el instanceof FormChoice) {
                getFormChoice((FormChoice)el, name, out);
            } else if (el instanceof FormBarCode) {
                getFormBarCode((FormBarCode)el, name, out);
            } else {
                out.startSection("Unknown field type for field " + name);
            }
            if (el.getDescription() != null) {
                out.println("Description:\t" + el.getDescription());
            }
            if (el.isReadOnly()) {
                out.println("Read-Only:\ttrue");
            }
            if (el.isRequired()) {
                out.println("Required:\ttrue");
            }
            if (!el.isSubmitted()) {
                out.println("Submitted:\tfalse");
            }
            for (int j=0;j<Event.ALL.length;j++) {
                PDFAction act = el.getAction(Event.ALL[j]);
                getAction(act, Event.ALL[j].toString(), out);
            }
            out.endSection();
        }
        out.endSection();
    }

    void getFormSignature(FormSignature sig, String name, Printer out) throws Exception {
        String[] certifications = { "No", "Yes: No Changes", "Yes: Modify Forms", "Yes: Modify Comments" };
        out.startSection("Signature \"" + name + "\"");
        if (sig.getState() == sig.STATE_BLANK) {
            out.println("State:\tBlank");
        } else {
            Calendar when = sig.getSignDate();
            out.println("Type:\t" + sig.getFilter()+(sig.getSignatureHandler() instanceof PKCS7SignatureHandler ? " / "+((PKCS7SignatureHandler)sig.getSignatureHandler()).getSubFilter() : ""));
            out.println("Handler:\t" + sig.getSignatureHandler().getHandlerName());
            if (keystore == null) {
                keystore = FormSignature.loadDefaultKeyStore();
            }
            Object[] o = getSignatureValidity(sig, keystore);
            String validity = (String)o[0];
            X509Certificate cert = (X509Certificate)o[1];
            if (when != null) { // Only null for non-standard SignatureHandlers
                String timestampvalidity = getSignatureTimeStampValidity(sig, keystore);
                dateformat.setTimeZone(when.getTimeZone());
                out.println("Signed:\t" + dateformat.format(when.getTime()) + " (" + timestampvalidity + ")");
                dateformat.setTimeZone(TimeZone.getDefault());
            }
            out.println("Certified:\t" + certifications[sig.getCertificationType()]);
            if (sig.getName() != null) {
                out.println("Signed by:\t" + sig.getName());
            }
            if (sig.getReason() != null) {
                out.println("Reason:\t" + sig.getReason());
            }
            if (sig.getLocation() != null) {
                out.println("Location:\t" + sig.getLocation());
            }
            out.println("Validity:\t" + validity);
            if (sig.getSignatureHandler() instanceof PKCS7SignatureHandler) {
                PKCS7SignatureHandler pkcs7 = (PKCS7SignatureHandler)sig.getSignatureHandler();
                out.println("Digest Algo:\t" + pkcs7.getHashAlgorithm());
                List<PKCS7SignatureHandler.ValidationInformation> ltv = pkcs7.getValidationInformation();
                if (ltv != null) {
                    for (PKCS7SignatureHandler.ValidationInformation vi : ltv) {
                        out.startSection("Long Term Validation");
                        out.println("When:\t" + (vi.getTime() == null ? null : vi.getTime().getTime()));
                        out.println("Original:\t" + vi.isInitial());
                        out.println("Complete:\t" + vi.isComplete(null));
                        out.endSection();
                    }
                }
            }
            if (cert != null) {
                out.startSection("Unverifiable Certificate");
                out.println(cert.toString());
                out.endSection();
            }
        }
    }

    /**
     * Verify the signature.
     *
     * Unfortunately we want to return two things here - the "validity" string
     * and the unverifiable X.509 certificate. We return them in an array...
     */
    private Object[] getSignatureValidity(FormSignature sig, KeyStore keystore) {
        boolean verified=false, verifiable=false, certsok=false;
        X509Certificate cert = null;
        Calendar when = sig.getSignDate();
        try {
            verified = sig.verify();
            verifiable = true;
            if (sig.getSignatureHandler() instanceof PKCS7SignatureHandler) {
                PKCS7SignatureHandler handler = (PKCS7SignatureHandler)sig.getSignatureHandler();
                X509Certificate[] certs = handler.getCertificates();
                cert = FormSignature.verifyCertificates(certs, keystore, null, when);
                certsok = cert == null;
            } else {
            }
        } catch (Exception e) {
            System.out.println("ERROR: Verifying certificate");
            e.printStackTrace();
        }
        String validity="";
        if (sig.getNumberOfRevisionsCovered() != pdf.getNumberOfRevisions()) {
            if (sig.getNumberOfRevisionsCovered() == 0) {
                validity = "Bad - signature doesn't cover an entire revision. The section that is covered is ";
            } else {
                validity = "Partial - covers the first " + sig.getNumberOfRevisionsCovered() + " of " + pdf.getNumberOfRevisions() + " revisions. Those covered are ";
            }
        }
        if (verified && verifiable && certsok) {
            validity += "Perfect (signature and certificates verified)";
        } else if (verified && verifiable && !certsok) {
            validity += "Good (signature verified, but not certificates)";
        } else if (verifiable) {
            validity += "Bad (signature not verified - document has been altered)";
        } else {
            validity += "Unknown (not a PKCS#7 Signature, unable to verify)";
        }
        return new Object[] { validity, cert };
    }

    /**
     * Verify the RFC3161 TimeStamp on the signature
     * if there is one (only on PKCS#7 sigs).
     */
    private String getSignatureTimeStampValidity(FormSignature sig, KeyStore keystore) {
        X509Certificate[] certs = null;
        if (sig.getSignatureHandler() instanceof PKCS7SignatureHandler) {
            try {
                certs = ((PKCS7SignatureHandler)sig.getSignatureHandler()).getTimeStampCertificates();
            } catch (Exception e) {
                System.out.println("ERROR: Verifying TimeStamp certificates");
                e.printStackTrace();
            }
        }

        String validity;
        if (certs != null) {
            Calendar when = sig.getSignDate();
            String issuer = null;
            X509Certificate badcert = null;
            try {
                issuer = FormSignature.getSubjectField(certs[0], "CN");
                badcert = FormSignature.verifyCertificates(certs, keystore, null, when);
            } catch (Exception e) {
                System.out.println("ERROR: Verifying certificate");
                e.printStackTrace();
            }
            if (badcert == null) {
                validity = "guaranteed by verified timestamp from \"" + issuer + "\")";
            } else {
                validity = "guaranteed by unknown timestamp from \"" + issuer + "\")";
            }
        } else {
            validity = "from computer clock";
        }
        return validity;
    }

    void getFormText(FormText text, String name, Printer out) throws Exception {
        out.startSection("Text \"" + name + "\"");
        out.println("Value:\t" + (text.getValue() == null ? null : fmt(text.getValue())));
        if (text.getDefaultValue() != null) {
            out.println("DefaultValue:\t" + fmt(text.getDefaultValue()));
        }
        if (text.getType() == FormText.TYPE_PASSWORD) {
            out.println("Type:\t\tPassword");
        }
        if (text.getType() == FormText.TYPE_BARCODE) {
            out.println("Type:\t\tXFA Barcode");
        }
        if (text.getType() == FormText.TYPE_MULTILINE) {
            out.println("Type:\t\tmulti-line");
        }
        if (text.getMaxLength() != 0) {
            out.println("Max-Length:\t" + text.getMaxLength());
        }
        if (text.getNumberOfCombs() != 0) {
            out.println("Combs:\t" + text.getNumberOfCombs());
        }
        if (text.isSpellCheck()) {
            out.println("SpellCheck:\ttrue");
        }
        if (text.isScrollable()) {
            out.println("Scrollable:\ttrue");
        }
    }

    void getFormButton(FormButton button, String name, Printer out) throws Exception {
        out.startSection("Button \"" + name + "\"");
        out.println("Buttons have no attributes");
    }

    void getFormRadioBox(FormElement box, String name, Printer out) throws Exception {
        out.startSection((box instanceof FormRadioButton ? "RadioButton" : "Checkbox") + " \"" + name + "\"");
        if (box.getValue() != null) {
            out.println("Value:\t" + box.getValue());
        }
        List<WidgetAnnotation> annots = box.getAnnotations();
        for (int i=0;i<annots.size();i++) {
            WidgetAnnotation w = annots.get(i);
            out.println(w.getValue());
        }
    }

    void getFormChoice(FormChoice choice, String name, Printer out) throws Exception {
        out.startSection("Choice \"" + name + "\"");
        int type = choice.getType();
        out.println("Type:\t" + (type == FormChoice.TYPE_COMBO ? "Combo" : type == FormChoice.TYPE_SCROLLABLE ? "Scrollable" : type == FormChoice.TYPE_MULTISCROLLABLE ? "Multi-choice Scrollable" : "DropDown"));
        if (choice.getValue() != null) {
            out.println("Value:\t" + choice.getValue());
        }

        Map<String,String> opts = choice.getOptions();
        for (Iterator<Map.Entry<String,String>> j=opts.entrySet().iterator();j.hasNext();) {
            Map.Entry<String,String> e = j.next();
            String key = e.getKey();
            String val = e.getValue();
            out.println(" * " + key + " = " + val);
        }
    }
    
    void getFormBarCode(FormBarCode barcode, String name, Printer out) throws Exception {
        out.startSection("Barcode \"" + name + "\"");
        out.println("Symbology:\t" + barcode.getSymbology() + " (with ECC=" + barcode.getECC() + " and X-unit of " + fmt.format(barcode.getSymbolSize()) + "mm" + (barcode.getCompression() != 0?", flate compressed":"") + ")");
        if (barcode.getValue() != null) {
            out.println("Value:\t" + fmt(barcode.getValue()));
        }
    }

    //-------------------------------------------------------------------------------------
    //
    // Remainder are utility methods and a subclass (Printer) for formatting the output
    //
    //-------------------------------------------------------------------------------------

    private static final DecimalFormat fmt = new DecimalFormat("#0.00");

    private static String fmt(double f) {
        if (f != f) {
            return "NaN";
        } else if (f == Double.POSITIVE_INFINITY) {
            return " + Inf";
        } else if (f == Double.POSITIVE_INFINITY) {
            return "-Inf";
        }  else {
            return f == (int)f ? Integer.toString((int)f) : fmt.format(f);
        }
    }

    private String fmt(String in) {
        StringBuffer st = new StringBuffer(in.length());
        for (int i=0;i<in.length();i++) {
            char c = in.charAt(i);
            if (c == '\n') {
                st.append("\\n");
            } else if (c == '\r') {
                st.append("\\r");
            } else if (c == '\t') {
                st.append("\\t");
            } else if ((c>=32 && c<=255) || unicode) {
                st.append(c);
            } else {
                String z = "0000" + Integer.toHexString(c);
                st.append("\\u" + z.substring(z.length()-4, z.length()));
            }
        }
        return st.toString();
    }

    private static String fmt(Shape s) {
        if (s instanceof Rectangle2D) {
            Rectangle2D r = (Rectangle2D)s;
            return "<rect x=" + fmt(r.getMinX()) + " y=" + fmt(r.getMinY()) + " w=" + fmt(r.getWidth()) + " h=" + fmt(r.getHeight()) + ">";
        } else if (s instanceof Ellipse2D) {
            Ellipse2D r = (Ellipse2D)s;
            return "<ellipse x=" + fmt(r.getMinX()) + " y=" + fmt(r.getMinY()) + " w=" + fmt(r.getWidth()) + " h=" + fmt(r.getHeight()) + ">";
        } else if (s instanceof Line2D) {
            Line2D l = (Line2D)s;
            return "<line x1=" + fmt(l.getX1()) + " y1=" + fmt(l.getY1()) + " x2=" + fmt(l.getX2()) + " y2=" + fmt(l.getY2()) + ">";
        } else {
            float[] f = new float[6];
            StringBuffer sb = new StringBuffer();
            for (PathIterator i = s.getPathIterator(null);!i.isDone();i.next()) {
                int t = i.currentSegment(f);
                if (t == i.SEG_MOVETO) {
                    sb.append(fmt(f[0]) + " " + fmt(f[1]) + " m ");
                } else if (t == i.SEG_LINETO) {
                    sb.append(fmt(f[0]) + " " + fmt(f[1]) + " l ");
                } else if (t == i.SEG_QUADTO) {
                    sb.append(fmt(f[0]) + " " + fmt(f[1]) + " " + fmt(f[2]) + " " + fmt(f[3]) + " q ");
                } else if (t == i.SEG_CUBICTO) {
                    sb.append(fmt(f[0]) + " " + fmt(f[1]) + " " + fmt(f[2]) + " " + fmt(f[3]) + " " + fmt(f[4]) + " " + fmt(f[5]) + " c ");
                } else {
                    sb.append(" h ");
                }
            }
            return sb.toString().trim();
        }
    }

    /** 
     * Utility class to format output nicely
     */
    static class Printer {
        final PrintWriter writer;
        private int printdepth, endprintdepth;
        private List<String> headers;
        private boolean nl = true;
        private String prefix = "";
        static final String LINE = "---------------------------------------------------------------";

        Printer(PrintWriter writer) {
            this.writer = writer;
            this.headers = new ArrayList<String>();
        }

        void startSection(String header) {
            headers.add(header);
        }

        void endSection() {
            if (printdepth == headers.size()) {
                if (printdepth == 1) {
                    writer.println();
                } else {
                    String header = headers.get(headers.size()-1);
                    prefix = prefix.substring(0, prefix.length()-2);
                    writer.println(prefix + "+" + LINE);
                    endprintdepth = printdepth;
                }
                printdepth--;
            }
            headers.remove(headers.size()-1);
            writer.flush();
        }

        void println(StringBuffer s) {
            println(s.toString());
        }

        void print(StringBuffer s) {
            print(s.toString());
        }

        void println(String s) {
            print(s + "\n");
        }

        void print(String s) {
            while (printdepth < headers.size()) {
                String header = headers.get(printdepth);
                if (printdepth == 0) {
                    writer.println("=> " + header);
                    writer.println("-" + LINE);
                } else {
                    if (printdepth+1 != endprintdepth) {
                        writer.println(prefix + "+" + LINE);
                    }
                    prefix += "| ";
                    writer.println(prefix + "=> " + header);
                }
                printdepth++;
            }
            if (nl) {
                writer.print(prefix);
            }
            writer.print(s);
            nl = s.endsWith("\n");
        }
    }
}
